# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
# based on work of:
# Copyright (c) 2018 Antonio Ospite <ao2@ao2.it>

from svgwrite.data.types import SVGAttribute

INKSCAPE_NAMESPACE = 'http://www.inkscape.org/namespaces/inkscape'
SODIPODI_NAMESPACE = 'http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd'
INKSCAPE_ATTRIBUTES = {
    'xmlns:inkscape': SVGAttribute('xmlns:inkscape',
                                   anim=False,
                                   types=[],
                                   const=frozenset([INKSCAPE_NAMESPACE])),
    'xmlns:sodipodi': SVGAttribute('xmlns:sodipodi',
                                   anim=False,
                                   types=[],
                                   const=frozenset([SODIPODI_NAMESPACE])),
    'inkscape:groupmode': SVGAttribute('inkscape:groupmode',
                                       anim=False,
                                       types=[],
                                       const=frozenset(['layer'])),
    'inkscape:label': SVGAttribute('inkscape:label',
                                   anim=False,
                                   types=frozenset(['string']),
                                   const=[]),
    'sodipodi:insensitive': SVGAttribute('sodipodi:insensitive',
                                         anim=False,
                                         types=[],
                                         const=frozenset(['true', 'false', '0', '1']))
}


def _setup_validator(validator):
    # setup already done?
    if 'xmlns:inkscape' in validator.attributes:
        return

    validator.attributes.update(INKSCAPE_ATTRIBUTES)
    elements = validator.elements

    # extend SVG attributes
    elements['svg'].valid_attributes = \
        {
            'xmlns:inkscape',
            'xmlns:sodipodi',
        } | elements['svg'].valid_attributes

    # extend group attributes
    elements['g'].valid_attributes = \
        {
            'inkscape:groupmode',
            'inkscape:label',
            'sodipodi:insensitive',
        } | elements['g'].valid_attributes


GROUP_MODE = 'inkscape:groupmode'
LABEL = 'inkscape:label'
INSENSITIVE = 'sodipodi:insensitive'


class Inkscape(object):
    """
    Extension to support SOME Inkscape features.

    """
    def __init__(self, drawing):
        self.svg = drawing
        _setup_validator(drawing.validator)
        drawing['xmlns:inkscape'] = INKSCAPE_NAMESPACE
        drawing['xmlns:sodipodi'] = SODIPODI_NAMESPACE

    def layer(self, label=None, locked=False, **kwargs):
        """
        Create new Inkscape layer.

        Args:
            label: layer name as string
            locked: when set to True, make objects at this layer unselectable

        """
        new_layer = self.svg.g(**kwargs)
        new_layer[GROUP_MODE] = 'layer'
        if label is not None:
            new_layer[LABEL] = label
        if locked:
            new_layer[INSENSITIVE] = 1
        return new_layer



"""
Extension to create and manipulate shapes
"""
# Copyright (c) 2019 Christof Hanke (christof.hanke@induhviduals.de)
# License: MIT License
import math


def ngon(num_corners, edge_length=None, radius=None, rotation=0.):
    """
    Returns the corners of a regular polygon as iterable of (x, y) tuples. The polygon size is determined by the
    `edge_length` or the `radius` argument. If both are given `edge_length` will be taken.

    Args:
        num_corners: count of polygon corners
        edge_length: length of polygon side
        radius: circum radius
        rotation: rotation angle in radians

    Returns: iterable of (x, y) tuples

    """
    if num_corners < 3:
        raise ValueError('Argument `num_corners` has to be greater than 2.')
    if edge_length is not None:
        radius = edge_length / 2 / math.sin(math.pi / num_corners)
    elif radius is not None:
        if radius <= 0.:
            raise ValueError('Argument `radius` has to be greater than 0.')
    else:
        raise ValueError('Argument `edge_length` or `radius` required.')

    delta = 2 * math.pi / num_corners
    angle = rotation
    for _ in range(num_corners):
        yield (radius * math.cos(angle), radius * math.sin(angle))
        angle += delta


def star(spikes, r1, r2, rotation=0.):
    """
    Create a star shape as iterable of (x, y) vertices.

    Argument `spikes` defines the count of star spikes, `r1` defines the radius of the "outer" vertices and `r2`
    defines the radius of the "inner" vertices, but this does not mean that `r1` has to greater than `r2`.

    Args:
        spikes: spike count
        r1: radius 1
        r2: radius 2
        rotation: rotation angle in radians

    Returns: iterable of (x, y) tuples

    """
    if spikes < 3:
        raise ValueError('Argument `spikes` has to be greater than 2.')
    if r1 <= 0.:
        raise ValueError('Argument `r1` has to be greater than 0.')
    if r2 <= 0.:
        raise ValueError('Argument `r2` has to be greater than 0.')

    corners1 = ngon(spikes, radius=r1, rotation=rotation)
    corners2 = ngon(spikes, radius=r2, rotation=math.pi/spikes+rotation)
    for s1, s2 in zip(corners1, corners2):
        yield s1
        yield s2


def translate(vertices, delta_x, delta_y):
    """
    Translates `vertices` about `delta_x` and `delta_y`

    Args:
         vertices: iterable of (x, y) tuples
         delta_x: translation in x axis
         delta_y: translation in y axis

    Returns: iterable of (x, y) tuples

    """
    for x, y in vertices:
        yield (x + delta_x, y + delta_y)


def scale(vertices, scale_x, scale_y):
    """
    Scales `vertices` about `scale_x` and `scale_y`

    Args:
         vertices: iterable of (x, y) tuples
         scale_x: scaling factor in x axis direction
         scale_y: scaling factor in y axis direction

    Returns: iterable of (x, y) tuples

    """
    for x, y in vertices:
        yield (x * scale_x, y * scale_y)


def rotate(vertices, delta):
    """
    Rotates `vertices` about `delta` degrees around the origin (0, 0).

    Args:
         vertices: iterable of (x, y) tuples
         delta: rotation angle in radians

    Returns: iterable of (x, y) tuples

    """
    for x, y in vertices:
        r = math.hypot(x, y)
        angle = math.atan2(y, x) + delta
        yield (r * math.cos(angle), r * math.sin(angle))


def centroid(vertices):
    """
    Returns the centroid of a series of `vertices`.

    """
    k, c_x, c_y = 0, 0, 0
    for x, y in vertices:
        c_x += x
        c_y += y
        k += 1
    return c_x / k, c_y / k



#!/usr/bin/env python
# coding:utf-8
# Authors:  mozman <me@mozman.at>, Florian Festi
# Purpose: svgparser using re module
# Created: 16.10.2010
# Copyright (c) 2010, Manfred Moitzi & 2020, Florian Festi
# License: MIT License

__all__ = ["is_valid_transferlist", "is_valid_pathdata", "is_valid_animation_timing"]

import re

event_names = [
    "focusin", "focusout", "activate", "click", "mousedown", "mouseup", "mouseover",
    "mousemove", "mouseout", "DOMSubtreeModified", "DOMNodeInserted", "DOMNodeRemoved",
    "DOMNodeRemovedFromDocument", "DOMNodeInsertedtoDocument", "DOMAttrModified",
    "DOMCharacterDataModified", "SVGLoad", "SVGUnload", "SVGAbort", "SVGError",
    "SVGResize", "SVGScroll", "SVGZoom", "beginEvent", "endEvent", "repeatEvent",
]

c = r"\s*[, ]\s*"
s = r"\s*[; ]\s*"
integer_constant = r"\d+"
exponent = r"([eE][+-]?\d+)"
nonnegative_number = fr"(\d+\.?\d*|\.\d+){exponent}?"
number = r"[+-]?" + nonnegative_number
flag = r"[01]"

comma_delimited_coordinates = fr"\s*{number}({c}{number})*\s*"
two_comma_delimited_numbers = fr"\s*{number}({c}{number}\s*)" "{1}"
four_comma_delimited_numbers = fr"\s*{number}\s*({c}{number}\s*)" "{3}"
six_comma_delimited_numbers = fr"\s*{number}\s*({c}{number}\s*)" "{5}"
comma_delimited_coordinate_pairs = fr"{two_comma_delimited_numbers}({c}{two_comma_delimited_numbers})*"


def is_valid(regex):
    reg = re.compile(regex)

    def f(term):
        return bool(reg.fullmatch(term))

    return f


def build_transferlist_parser():
    matrix = fr"matrix\s*\(\s*{six_comma_delimited_numbers}\s*\)"
    translate = fr"translate\s*\(\s*{number}({c}{number})?\s*\)"
    scale = fr"scale\s*\(\s*{number}({c}{number})?\s*\)"
    rotate = fr"rotate\s*\(\s*{number}({c}{number}{c}{number})?\s*\)"
    skewX = fr"skewX\s*\(\s*{number}\s*\)"
    skewY = fr"skewY\s*\(\s*{number}\s*\)"

    tl_re = "|".join((fr"(\s*{cmd}\s*)" for cmd in (
        matrix, translate, scale, rotate, skewX, skewY)))
    return fr"({tl_re})({c}({tl_re}))*"


is_valid_transferlist = is_valid(build_transferlist_parser())


def build_pathdata_parser():
    moveto = fr"[mM]\s*{comma_delimited_coordinate_pairs}"
    lineto = fr"[lL]\s*{comma_delimited_coordinate_pairs}"
    horizontal_lineto = fr"[hH]\s*{comma_delimited_coordinates}"
    vertical_lineto = fr"[vV]\s*{comma_delimited_coordinates}"
    curveto = fr"[cC]\s*({six_comma_delimited_numbers})({c}{six_comma_delimited_numbers})*"
    smooth_curveto = fr"[sS]{four_comma_delimited_numbers}({c}{four_comma_delimited_numbers})*"
    quadratic_bezier_curveto = fr"[qQ]{four_comma_delimited_numbers}({c}{four_comma_delimited_numbers})*"
    smooth_quadratic_bezier_curveto = fr"[tT]\s*{comma_delimited_coordinate_pairs}"

    elliptical_arc_argument = fr"{c}".join((
        fr"{nonnegative_number}",
        fr"{nonnegative_number}",
        fr"{number}",
        fr"{flag}",
        fr"{flag}",
        fr"{number}",
        fr"{number}",))
    elliptical_arc_argument = r"\s*" + elliptical_arc_argument + r"\s*"
    elliptical_arc = fr"[aA]({elliptical_arc_argument})({c}{elliptical_arc_argument})*"

    drawto_command = "|".join((fr"(\s*{cmd}\s*)" for cmd in (
        moveto, lineto, horizontal_lineto, vertical_lineto, "[zZ]",
        curveto, smooth_curveto, quadratic_bezier_curveto,
        smooth_quadratic_bezier_curveto, elliptical_arc)))

    return f"{moveto}({drawto_command})*"


is_valid_pathdata = is_valid(build_pathdata_parser())

digit2 = r"\d{2}"
digit4 = r"\d{4}"
seconds = fr"\d+(\.\d+)?"
seconds2 = fr"{digit2}(\.\d+)?"
metric = "(h|min|s|ms)"


def clock_val_re():
    timecount_val = fr"{seconds}\s*({metric})?"
    clock_val = fr"{digit2}:({digit2}:)?{seconds2}"
    return fr"({timecount_val}|{clock_val})"


def wall_clock_val_re():
    hhmmss = fr"{digit2}:{digit2}(:{seconds2})?"
    walltime = fr"{hhmmss}(Z|[+-]?{digit2}:{digit2})?"
    date = fr"{digit4}-{digit2}-{digit2}"
    datetime = fr"{date}(T{walltime})?"
    return "(" + "|".join((walltime, datetime)) + ")"


def build_animation_timing_parser():
    clock_val = clock_val_re()
    wallclock_val = wall_clock_val_re()

    event_ref = "(" + "|".join(event_names) + ")"
    id_value = "#?[-_a-zA-Z0-9]+"

    wallclock_sync_value = fr"wallclock\(\s*{wallclock_val}\s*\)"
    accesskey_value = fr"accessKey\(\s*[a-zA-Z]\s*\)\s*([+-]?{clock_val})?"
    repeat_value = fr"({id_value}\.)?repeat\s*\(\s*\d+\s*\)\s*([+-?]{clock_val})?"
    event_value = fr"({id_value}\.)?{event_ref}([+-]?{clock_val})?"
    offset_value = fr"[-+]?{clock_val}"
    syncbase_value = fr"{id_value}\.(begin|end)({offset_value})?"
    begin_value = "(" + "|".join((f"({reg})" for reg in (
        offset_value, syncbase_value, event_value, repeat_value,
        accesskey_value, wallclock_sync_value, "indefinite"))) + ")"
    return fr"{begin_value}({s}{begin_value})*"


is_valid_animation_timing = is_valid(build_animation_timing_parser())



#!/usr/bin/env python
#coding:utf-8
# Author:  mozman --<mozman@gmx.at>
# Purpose: solidColor paint serve (Tiny 1.2 profile)
# Created: 26.10.2016
# Copyright (C) 2016, Manfred Moitzi
# License: MIT License

from svgwrite.base import BaseElement
from svgwrite.mixins import XLink


class SolidColor(BaseElement, XLink):
    """
    The `solidColor` element is a paint server that provides a single color with opacity.
    It can be referenced like the other paint servers (i.e. gradients).
    The `color` parameter specifies the color that shall be used for this `solidColor` element.
    The keyword ``"currentColor"`` can be specified in the same manner as within a <paint> specification for the `fill`
    and `stroke` properties. The `opacity` parameter defines the opacity of the `solidColor`.
    """
    elementname = 'solidColor'

    def __init__(self, color="currentColor", opacity=None, **extra):
        """
        :param color: solid color like the other paint servers (i.e. gradients).
        :param float opacity: opacity of the solid color in the range `0.0` (fully transparent) to `1.0` (fully opaque)

        """
        super(SolidColor, self).__init__(**extra)
        if self.profile != 'tiny':
            raise TypeError("Paint server 'solidColor' requires the Tiny SVG profile.")
        self['solid-color'] = color
        if opacity is not None:
            self['solid-opacity'] = opacity

        if self.debug:
            self.validator.check_all_svg_attribute_values(self.elementname, self.attribs)

    def get_paint_server(self, default='none'):
        """ Returns the <FuncIRI> of the gradient. """
        return "%s %s" % (self.get_funciri(), default)
